use std::{collections::HashMap, ops::Range};

use cgmath::{prelude::*, Vector3};

const ADAPTIVE: bool = false;

pub fn generate_mesh<F, FN>(
    f: F,
    f_normal: FN,
    xs: Range<i32>,
    ys: Range<i32>,
    zs: Range<i32>,
) -> (Vec<Vector3<f32>>, Vec<[usize; 3]>)
where
    F: Fn(i32, i32, i32) -> f32,
    FN: Fn(f32, f32, f32) -> Vector3<f32>,
{
    let mut verts = vec![];
    let mut idx = HashMap::new();

    for x in &mut xs.clone() {
        for y in &mut ys.clone() {
            for z in &mut zs.clone() {
                if let Some(vert) = find_best_vertex(&f, &f_normal, x, y, z) {
                    idx.insert((x, y, z), verts.len());
                    verts.push(vert);
                }
            }
        }
    }

    let mut faces = vec![];

    for x in &mut xs.clone() {
        for y in &mut ys.clone() {
            for z in &mut zs.clone() {
                if x > xs.start && y > ys.start {
                    let s1 = f(x, y, z + 0) > 0.0;
                    let s2 = f(x, y, z + 1) > 0.0;
                    if s1 != s2 {
                        let mut a = [
                            idx[&(x - 1, y - 1, z)],
                            idx[&(x - 0, y - 1, z)],
                            idx[&(x - 0, y - 0, z)],
                        ];
                        let mut b = [
                            idx[&(x - 1, y - 1, z)],
                            idx[&(x - 0, y - 0, z)],
                            idx[&(x - 1, y - 0, z)],
                        ];
                        if s2 {
                            a.reverse();
                            b.reverse();
                        }
                        faces.push(a);
                        faces.push(b);
                    }
                }
                if x > xs.start && z > zs.start {
                    let s1 = f(x, y + 0, z) > 0.0;
                    let s2 = f(x, y + 1, z) > 0.0;
                    if s1 != s2 {
                        let mut a = [
                            idx[&(x - 1, y, z - 1)],
                            idx[&(x - 0, y, z - 1)],
                            idx[&(x - 0, y, z - 0)],
                        ];
                        let mut b = [
                            idx[&(x - 1, y, z - 1)],
                            idx[&(x - 0, y, z - 0)],
                            idx[&(x - 1, y, z - 0)],
                        ];
                        if s1 {
                            a.reverse();
                            b.reverse();
                        }
                        faces.push(a);
                        faces.push(b);
                    }
                }
                if y > ys.start && z > zs.start {
                    let s1 = f(x + 0, y, z) > 0.0;
                    let s2 = f(x + 1, y, z) > 0.0;
                    if s1 != s2 {
                        let mut a = [
                            idx[&(x, y - 1, z - 1)],
                            idx[&(x, y - 0, z - 1)],
                            idx[&(x, y - 0, z - 0)],
                        ];
                        let mut b = [
                            idx[&(x, y - 1, z - 1)],
                            idx[&(x, y - 0, z - 0)],
                            idx[&(x, y - 1, z - 0)],
                        ];
                        if s2 {
                            a.reverse();
                            b.reverse();
                        }
                        faces.push(a);
                        faces.push(b);
                    }
                }
            }
        }
    }

    (verts, faces)
}

fn find_best_vertex<F, FN>(f: F, f_normal: FN, x: i32, y: i32, z: i32) -> Option<Vector3<f32>>
where
    F: Fn(i32, i32, i32) -> f32,
    FN: Fn(f32, f32, f32) -> Vector3<f32>,
{
    if !ADAPTIVE {
        return Some(Vector3::new(x as f32 + 0.5, y as f32 + 0.5, z as f32 + 0.5));
    }

    let mut v = [[[0.0; 2]; 2]; 2];
    for dx in 0..2 {
        for dy in 0..2 {
            for dz in 0..2 {
                v[dx][dy][dz] = f(x + dx as i32, y + dy as i32, z + dz as i32);
            }
        }
    }

    let mut changes = vec![];
    for dx in 0..2 {
        for dy in 0..2 {
            if (v[dx][dy][0] > 0.0) != (v[dx][dy][1] > 0.0) {
                changes.push(Vector3::new(
                    x as f32 + dx as f32,
                    y as f32 + dy as f32,
                    z as f32 + adapt(v[dx][dy][0], v[dx][dy][1]),
                ))
            }
        }
    }
    for dx in 0..2 {
        for dz in 0..2 {
            if (v[dx][0][dz] > 0.0) != (v[dx][1][dz] > 0.0) {
                changes.push(Vector3::new(
                    x as f32 + dx as f32,
                    y as f32 + adapt(v[dx][0][dz], v[dx][1][dz]),
                    z as f32 + dz as f32,
                ))
            }
        }
    }
    for dy in 0..2 {
        for dz in 0..2 {
            if (v[0][dy][dz] > 0.0) != (v[1][dy][dz] > 0.0) {
                changes.push(Vector3::new(
                    x as f32 + adapt(v[0][dy][dz], v[1][dy][dz]),
                    y as f32 + dy as f32,
                    z as f32 + dz as f32,
                ))
            }
        }
    }

    if changes.len() <= 1 {
        return None;
    }

    let normals: Vec<_> = changes
        .iter()
        .map(|v| f_normal(v.x, v.y, v.z))
        .collect();

    Some(qef(x, y, z, &changes, &normals))
}

fn qef(x: i32, y: i32, z: i32, pos: &[Vector3<f32>], normals: &[Vector3<f32>]) -> Vector3<f32> {
    todo!()
}

fn adapt(a: f32, b: f32) -> f32 {
    if ADAPTIVE {
        (-a) / (b - a)
    } else {
        0.5
    }
}

pub fn normal_from_function<F>(f: F, d: f32) -> impl Fn(f32, f32, f32) -> Vector3<f32>
where
    F: Fn(f32, f32, f32) -> f32,
{
    move |x, y, z| {
        Vector3::new(
            (f(x + d, y, z) - f(x - d, y, z)) / 2.0 / d,
            (f(x, y + d, z) - f(x, y - d, z)) / 2.0 / d,
            (f(x, y, z + d) - f(x, y, z - d)) / 2.0 / d,
        )
    }
}
