use bracket_noise::prelude::*;
use cgmath::{prelude::*, AbsDiffEq, Vector3};
use rayon::prelude::*;

use crate::model::Vertex;

pub mod dual_contour;
pub mod marching_cubes;

#[repr(C)]
#[derive(Copy, Clone, Debug, bytemuck::Pod, bytemuck::Zeroable)]
pub struct TerrainVertex {
    position: [f32; 3],
    normal: [f32; 3],
}

impl Vertex for TerrainVertex {
    fn desc<'a>() -> wgpu::VertexBufferLayout<'a> {
        use std::mem;
        wgpu::VertexBufferLayout {
            array_stride: mem::size_of::<TerrainVertex>() as wgpu::BufferAddress,
            step_mode: wgpu::InputStepMode::Vertex,
            // attributes: &vertex_attr_array![ 0 => Float32x3, 1 => Float32x3 ],
            attributes: &[
                wgpu::VertexAttribute {
                    offset: 0,
                    shader_location: 0,
                    format: wgpu::VertexFormat::Float32x3,
                },
                wgpu::VertexAttribute {
                    offset: mem::size_of::<[f32; 3]>() as wgpu::BufferAddress,
                    shader_location: 1,
                    format: wgpu::VertexFormat::Float32x3,
                },
            ],
        }
    }
}

fn generate_terrain_voxels<const SIZE: usize>(
    seed: u64,
    chunk_coords: Vector3<i32>,
) -> Box<[[[f32; SIZE]; SIZE]; SIZE]> {
    let mut noise = FastNoise::new();
    let chunk_coords = chunk_coords * (SIZE - 1) as i32;

    noise.set_seed(seed);
    noise.set_frequency(0.02);

    let mut heightmap = Box::new([[0.0; SIZE]; SIZE]);
    heightmap.par_iter_mut().enumerate().for_each(|(x, v)| {
        v.par_iter_mut().enumerate().for_each(|(y, v)| {
            *v = (noise.get_noise(
                chunk_coords.x as f32 + x as f32,
                chunk_coords.z as f32 + y as f32,
            ) + 1.0)
                / 2.0
                * SIZE as f32
        })
    });

    noise.set_seed(seed + 1);
    noise.set_frequency(0.03);
    let mut chunk_values = Box::new([[[0.0; SIZE]; SIZE]; SIZE]);
    chunk_values.par_iter_mut().enumerate().for_each(|(x, v)| {
        v.par_iter_mut().enumerate().for_each(|(y, v)| {
            v.par_iter_mut().enumerate().for_each(|(z, v)| {
                *v = heightmap[x][z] - y as f32 - chunk_coords.y as f32;
                *v += 8.0
                    * noise.get_noise3d(
                        (chunk_coords.x + x as i32) as f32,
                        (chunk_coords.y + y as i32) as f32,
                        (chunk_coords.z + z as i32) as f32,
                    );
            })
        })
    });

    chunk_values
}

fn optimize_mesh(mesh: Vec<TerrainVertex>) -> (Vec<TerrainVertex>, Vec<usize>) {
    fn equal(a: &[f32; 3], b: &[f32; 3]) -> bool {
        a[0].abs_diff_eq(&b[0], 0.001)
            && a[1].abs_diff_eq(&b[1], 0.001)
            && a[2].abs_diff_eq(&b[2], 0.001)
    }

    let mut indices: Vec<usize> = vec![];
    let mut vertices: Vec<TerrainVertex> = vec![];

    // This scales incredibly poorly, larger meshes ends up taking multiple seconds
    for vert in mesh {
        if let Some((i, _v)) = vertices
            .iter()
            .enumerate()
            .find(|(_i, v)| equal(&vert.position, &v.position))
        {
            indices.push(i);
        } else {
            indices.push(vertices.len());
            vertices.push(vert);
        }
    }

    (vertices, indices)
}

#[cfg(test)]
mod tests {
    use super::*;
    #[test]
    fn marching_cubes_test() {
        const SIZE: i32 = 8;
        const HEIGHT: i32 = 3;
        let time_total = std::time::Instant::now();
        let chunks: Vec<(Vec<TerrainVertex>, Vec<usize>, Vector3<i32>)> = (0..SIZE.pow(2) * HEIGHT)
            .into_par_iter()
            .map(|i| {
                let pos = Vector3 {
                    x: i % SIZE - SIZE / 2,
                    y: i / SIZE.pow(2) - HEIGHT / 2,
                    z: (i / SIZE) % SIZE - SIZE / 2,
                };
                let time_vox = std::time::Instant::now();
                let voxels = generate_terrain_voxels::<32>(0, pos);
                let time_vox = std::time::Instant::now() - time_vox;

                let time_vert = std::time::Instant::now();
                let vertices = marching_cubes::generate_mesh(&voxels);
                let time_vert = std::time::Instant::now() - time_vert;

                let time_indexing = std::time::Instant::now();
                let (verts, idx) = optimize_mesh(vertices);
                let time_indexing = std::time::Instant::now() - time_indexing;
                eprintln!(
                    "Timing for chunk: {:?}, Voxels {}s, Vertices {}s, Indexing {}s",
                    pos,
                    time_vox.as_secs_f64(),
                    time_vert.as_secs_f64(),
                    time_indexing.as_secs_f64()
                );
                (verts, idx, pos)
            })
            .collect();
        let time_total = std::time::Instant::now() - time_total;
        eprintln!(
            "Total time taken for {} chunks: {}s",
            SIZE.pow(2) * HEIGHT,
            time_total.as_secs_f32()
        );

        let mut vert = String::new();
        let mut norm = String::new();
        let mut face = String::new();

        let mut id_offset = 1;
        for (verts, idx, pos) in chunks {
            let pos = Vector3::new(
                (pos.x * 31) as f32,
                (pos.y * 31) as f32,
                (pos.z * 31) as f32,
            );
            for v in &verts {
                let [x, y, z] = v.position;
                assert!(!x.is_nan() && !y.is_nan() && !z.is_nan(), "vertex was NaN!");
                vert.push_str(&format!("v {} {} {}\n", x + pos.x, y + pos.y, z + pos.z));
                // let [x, y, z] = v.normal;
                // assert!(!x.is_nan() && !y.is_nan() && !z.is_nan(), "normal was NaN!");
                // norm.push_str(&format!("vn {} {} {}\n", x, y, z));
            }

            for i in idx.chunks(3) {
                face.push_str(&format!(
                    "f {} {} {}\n",
                    id_offset + i[0],
                    id_offset + i[1],
                    id_offset + i[2]
                ));
            }
            id_offset += verts.len()
        }
        print!("s 1\n{}{}{}", vert, norm, face);
    }

    #[test]
    fn dual_contour_test() {
        const CHUNK: usize = 32;
        const SIZE: i32 = 1;
        const HEIGHT: i32 = 1;
        let time_total = std::time::Instant::now();
        let chunks: Vec<(Vec<Vector3<f32>>, Vec<[usize; 3]>, Vector3<i32>)> = (0..SIZE.pow(2)
            * HEIGHT)
            .into_par_iter()
            .map(|i| {
                let pos = Vector3 {
                    x: i % SIZE - SIZE / 2,
                    y: i / SIZE.pow(2) - HEIGHT / 2,
                    z: (i / SIZE) % SIZE - SIZE / 2,
                };
                let time_vox = std::time::Instant::now();
                let voxels = generate_terrain_voxels::<CHUNK>(0, pos);
                let time_vox = std::time::Instant::now() - time_vox;

                let time_vert = std::time::Instant::now();
                let (verts, faces) = dual_contour::generate_mesh(
                    |x, y, z| voxels[x as usize][y as usize][z as usize],
                    |_, _, _| Vector3::new(0.0, 0.0, 0.0),
                    0..SIZE * 31,
                    0..SIZE * 31,
                    0..SIZE * 31,
                );
                let time_vert = std::time::Instant::now() - time_vert;

                eprintln!(
                    "Timing for chunk: {:?}, Voxels {}s, Vertices {}s",
                    pos,
                    time_vox.as_secs_f64(),
                    time_vert.as_secs_f64(),
                );
                (verts, faces, pos)
            })
            .collect();
        let time_total = std::time::Instant::now() - time_total;
        eprintln!(
            "Total time taken for {} chunks: {}s",
            SIZE.pow(2) * HEIGHT,
            time_total.as_secs_f32()
        );

        let mut vert = String::new();
        let mut norm = String::new();
        let mut face = String::new();

        let mut id_offset = 1;
        for (verts, faces, pos) in chunks {
            let pos = Vector3::new(
                (pos.x * 31) as f32,
                (pos.y * 31) as f32,
                (pos.z * 31) as f32,
            );
            for v in &verts {
                assert!(
                    !v.x.is_nan() && !v.y.is_nan() && !v.z.is_nan(),
                    "vertex was NaN!"
                );
                vert.push_str(&format!(
                    "v {} {} {}\n",
                    v.x + pos.x,
                    v.y + pos.y,
                    v.z + pos.z
                ));
                // let [x, y, z] = v.normal;
                // assert!(!x.is_nan() && !y.is_nan() && !z.is_nan(), "normal was NaN!");
                // norm.push_str(&format!("vn {} {} {}\n", x, y, z));
            }

            for f in faces {
                face.push_str(&format!(
                    "f {} {} {}\n",
                    id_offset + f[0],
                    id_offset + f[1],
                    id_offset + f[2]
                ));
            }
            id_offset += verts.len()
        }
        print!("s 1\n{}{}{}", vert, norm, face);
    }
}
